from .base_node import BaseNode

import numpy as np
import pandas as pd


class DataHolderNode(BaseNode):
    
    def __init__(self, graph_uid, graph_alias, node_uid, value=None, verbose=False):
        super(DataHolderNode, self).__init__(graph_uid, graph_alias, node_uid, 'data_holder', verbose)

        self.value = value
        
    def get(self):
        return self.value
    
    def set_value(self, value):
        self.value = value
    
    def has_value(self):
        return self.value is not None

    def get_persisted_data_dim_as_str(self):
        """Currently supports dimensionality from:

        numpy ndarray
        pandas dataframe
        pyspark dataframe

        The dimensionality of other types of data defaults to "(0)"
        """
        if not self.has_value:
            raise ValueError("There is no value!")

        if hasattr(self.value, "rdd"):
            row_cnt = self.get().persist().count()
            col_cnt = len(self.get().columns)
            dim = (row_cnt, col_cnt)

        elif isinstance(self.value, np.ndarray):
            dim = self.value.shape

        elif isinstance(self.value, pd.DataFrame):
            dim = self.value.shape

        else:
            return("(1, )")

        return str(dim)
        
    def __del__(self):
        if self.verbose:
            print('{} released!'.format(self.node_uid))



from .base_node import BaseNode


class OperationNode(BaseNode):
    
    def __init__(self, graph_uid, graph_alias, node_uid, function, n_out, verbose=False, alias=None, graph_dict=None):
        super(OperationNode, self).__init__(graph_uid, graph_alias, node_uid, 'operation', verbose, alias or function.__name__)
        
        self.function = function
        self.n_out = n_out
        self.is_active = False

        self.graph_dict = graph_dict
    
    def is_activated(self):
        return self.is_active
    
    def activate(self):
        
        if self.is_active:
            return
        
        if self.verbose:
            print('{} activated!'.format(self.node_uid))
            
        self.is_active = True
        
    def deactivate(self):
        
        if not self.is_active:
            return
        
        if self.verbose:
            print('{} deactivated!'.format(self.node_uid))
            
        self.is_active = False
        
    def run(self):
        """run method will do four things with respect to the current op node
        1. it will get values from the parent data node(s)
        2. then, it will run its func and obtain its output value(s)
        3. then, it will set the value of its child data node(s) with the output value(s) (plural if n_out>1)
        4. then, it will try to release the memory from its parent data nodes
        """
        
        # these strong references will be destroyed once we leave this scope
        parent_data_nodes_values = [parent_data_node_weak_ref().get() 
                                    for parent_data_node_weak_ref 
                                    in self.parent_node_weak_refs]

        if self.verbose:
            print('running {}'.format(self.node_uid))

        output_values = self.function(*parent_data_nodes_values)
        
        if self.n_out > 1:
            for i, output_value in enumerate(output_values):
                self.child_node_weak_refs[i]().set_value(output_value)
        else:
            # if the method of current op node has no return statement
            if len(self.child_node_weak_refs) == 0:
                pass
            else:
                self.child_node_weak_refs[0]().set_value(output_values)
        
        # the immediate parent data nodes
        # if any of these are needed by their child op node other than this one, 
        # we don't release them
        for parent_data_node_weak_ref in self.parent_node_weak_refs:
            
            # assume we will need to release this parent data node until proven otherwise
            release_parent_data_node = True
            
            # if the parent node is persisted, no need to check for its children op nodes
            if parent_data_node_weak_ref().is_persisted():
                continue
            
            # checking the child op nodes of the current parent data node
            for child_op_node_weak_ref in parent_data_node_weak_ref().get_child_node_weak_refs():
                
                # if one of the child op node of the parent data node is not activated, 
                # this op node gives us no reason to overturn the prejudice for releasing.
                # also, we do not care about the status of data nodes under this op node
                if not child_op_node_weak_ref().is_activated():
                    continue
                
                # now traversing down to the data nodes of the op node
                # if any of the data nodes are not filled, we need the current data node
                # to feed into that op node, so we do not want to release the memory.
                # having even one such op node + data node is enough reason to persist the 
                # parent data node (from outer most loop)
                for child_output_data_node_weak_ref in child_op_node_weak_ref().get_child_node_weak_refs():
                    
                    # if any child data node value of active op node is missing, we need the current
                    # parent data node to stay persisted
                    if not child_output_data_node_weak_ref().has_value():
                        
                        # for op node with one output data node, this weakref is referring to the
                        # output node it just computed for
                        if self.verbose:
                            print('{} still needed at {}'.format(parent_data_node_weak_ref().get_node_uid(), 
                                                                 child_op_node_weak_ref().get_node_uid()))

                        # at this point, we don't need to check any other child op node of 
                        # this parent data node.
                        # however, we may need to check the next parent_data_node_weak_ref
                        release_parent_data_node = False
                        break
                
                # if we cannot release this parent data node, no need to check other op nodes
                # just go to the next parent data node
                if not release_parent_data_node:
                    break
            
            # if the prejudice for releasing has not been overturned we will release memory
            if release_parent_data_node:
                parent_data_node_weak_ref().release_memory()
                
        self.deactivate()



import weakref
import copy
import os
import ast
import inspect
import textwrap

from graphviz import Digraph


class ExtendedRef(weakref.ref):

    def get(self):
        return self().get()
    
    def get_node_uid(self):
        return self().get_node_uid()
    
    def release_memory(self):
        return self().release_memory()

    def has_value(self):
        return self().has_value()
    
def contains_return_statement(func):
    func = textwrap.dedent(inspect.getsource(func))
    func_source_tree = ast.walk(ast.parse(func))
    return any(isinstance(node, ast.Return) for node in func_source_tree)

def get_rank(node_properties_dict):
    node_graph_attributes_dict = node_properties_dict['attributes']
    rank = node_graph_attributes_dict['rank']
    return rank

def get_type(node_properties_dict):
    node_type = node_properties_dict['type']
    return node_type

def view_full(graph_dict, graph_attributes, verbose, current_graph_uid):
    
    ranks = set()

    for node_properties_dict in graph_dict.values():
        
        rank = get_rank(node_properties_dict)
        if rank is not None:
            ranks.add(rank)

    ranks = list(ranks)
    ranks.sort()
    
    subgraphs = []

    for rank in ranks:
        subgraph = {k: v for k, v in graph_dict.items() if rank == get_rank(v)}
        subgraphs.append(subgraph)
        
    ranked_subgraphs = subgraphs[0:-1]
    unranked_subgraph = subgraphs[-1]

    graph = Digraph()
    graph.attr(splines='true', overlap='false', ranksep=graph_attributes['graph_ranksep'])

    for ranked_subgraph in ranked_subgraphs:

        with graph.subgraph() as subg:

            subg.attr(rank='same')

            for k, v in ranked_subgraph.items():

                if v['type'] != 'operation':
                    continue

                label = v['node_uid'] if verbose else v['alias']

                if v['attributes']['shape'] is not None:
                    shape = str(v['attributes']['shape'])
                else:
                    shape = graph_attributes['op_node_shape']

                if v['attributes']['color'] is not None:
                    color = str(v['attributes']['color'])
                else:
                    color = graph_attributes['op_node_color']

                if v['attributes']['fontsize'] is not None:
                    fontsize = str(v['attributes']['fontsize'])
                else:
                    fontsize = graph_attributes['op_node_fontsize']

                if v['attributes']['shapesize'] is not None:
                    shapesize = str(v['attributes']['shapesize'])
                else:
                    shapesize = '0.0'
                            
                subg.node(
                    k, 
                    label=label, 
                    shape=shape, 
                    fontsize=fontsize, 
                    height=shapesize, width=shapesize, fillcolor=color, style='filled')

                for child in v['children']:
                    label = graph_dict[child]['node_uid'] if verbose else graph_dict[child]['alias']
                    graph.node(
                        child, 
                        label=label,
                        shape=graph_attributes['data_node_shape'], 
                        fontsize=graph_attributes['data_node_fontsize'], 
                        height='0.0', width='0.0')
                    graph.edge(k, child)

                for parent in v['parents']:
                    label = graph_dict[parent]['node_uid'] if verbose else graph_dict[parent]['alias']
                    graph.node(
                        parent, 
                        label=label,
                        shape=graph_attributes['data_node_shape'], 
                        fontsize=graph_attributes['data_node_fontsize'], 
                        height='0.0', width='0.0')
                    graph.edge(parent, k)

    for k, v in unranked_subgraph.items():

        if v['type'] != 'operation':
            continue
        
        # determine the op node shape here

        label = v['node_uid'] if verbose else v['alias']

        if v['attributes']['shape'] is not None:
            shape = str(v['attributes']['shape'])
        else:
            shape = graph_attributes['op_node_shape']

        if v['attributes']['color'] is not None:
            color = str(v['attributes']['color'])
        else:
            color = graph_attributes['op_node_color']

        if v['attributes']['fontsize'] is not None:
            fontsize = str(v['attributes']['fontsize'])
        else:
            fontsize = graph_attributes['op_node_fontsize']

        if v['attributes']['shapesize'] is not None:
            shapesize = str(v['attributes']['shapesize'])
        else:
            shapesize = '0.0'

        # we added the condition to check that this op node belongs to the same graph
        # this is so that we keep the graph_node_shape shape in the multi graph setting
        # we do not add this condition for data node, which means we will see the 
        # persist node shape shift when viewed with summary=False
        if current_graph_uid==v['graph_uid'] and graph_attributes['persist_record_shape'] and v['is_persisted']:
            shape = 'record'
            data_dim = v['data_dim']
            label = "{{{}|{}}}".format(label, data_dim)

        graph.node(
            k, 
            label=label, 
            shape=shape, 
            fontsize=fontsize, 
            height=shapesize, width=shapesize, fillcolor=color, style='filled')

        for child in v['children']:

            # determine the data node shape here

            label = graph_dict[child]['node_uid'] if verbose else graph_dict[child]['alias']


            if graph_attributes['persist_record_shape'] and graph_dict[child]['is_persisted']:
                shape = 'record'
                data_dim = graph_dict[child]['data_dim']
                label = "{{{}|{}}}".format(label, data_dim)

            else:
                shape = graph_attributes['data_node_shape']

            graph.node(
                child,
                label=label,
                shape=shape, 
                fontsize=graph_attributes['data_node_fontsize'], 
                height='0.0', width='0.0', fillcolor=color)
            graph.edge(k, child)

        for parent in v['parents']:

            label = graph_dict[parent]['node_uid'] if verbose else graph_dict[parent]['alias']

            if graph_attributes['persist_record_shape'] and graph_dict[parent]['is_persisted']:
                shape = 'record'
                data_dim = graph_dict[parent]['data_dim']
                label = "{{{}|{}}}".format(label, data_dim)

            else:
                shape = graph_attributes['data_node_shape']

            graph.node(
                parent, 
                label=label,
                shape=shape, 
                fontsize=graph_attributes['data_node_fontsize'], 
                height='0.0', width='0.0', fillcolor=color)
            graph.edge(parent, k)

    return graph

def view_summary(graph_dict, graph_attributes, verbose, current_graph_uid):
    
    op_subgraphs = {k: v for k, v in graph_dict.items() if 'operation' == v['type']}
    data_subgraphs = {k: v for k, v in graph_dict.items() if 'data' == v['type']}

    op_graph_dict = copy.deepcopy(op_subgraphs)

    for k, v in op_graph_dict.items():
        v['children'] = []
        v['parents'] = []

    for k, v in op_subgraphs.items():

        for child_data_node_uid in v['children']:

            data_node_prop_dict = data_subgraphs[child_data_node_uid]

            is_persisted = data_node_prop_dict['is_persisted']
            data_dim = data_node_prop_dict['data_dim']
            
            op_graph_dict[k]['is_persisted'] = is_persisted
            op_graph_dict[k]['data_dim'] = data_dim

            for child_op_node_uid in data_node_prop_dict['children']:

                if child_op_node_uid is not None:

                    op_graph_dict[k]['children'].append(child_op_node_uid)

    return view_full(op_graph_dict, graph_attributes, verbose, current_graph_uid)

def save_graph_image(graph, dirpath=None, filename=None, fileformat=None):
    
    graph.format = 'png' or fileformat
    try:
        dirpath = os.getcwd() or dirpath
        filename = 'digraph' or filename
        filename = filename
        graph.view(filename=filename, directory=dirpath)
    except FileNotFoundError:
        pass
    finally:
        img_filepath = os.path.join(dirpath, filename + '.' + graph.format)
        dot_filepath = os.path.join(dirpath, filename)
        if not os.path.exists(img_filepath):
            raise FileNotFoundError('{} image file not found!'.format(img_filepath))
        if os.path.exists(dot_filepath):
            os.remove(dot_filepath)
    
    return img_filepath
    
def _recursive_topological_sort(node_uid, graph_dict, stack, visited_flags):
    
    # we have visited this node
    visited_flags[node_uid] = True
    
    for child_node_uid in graph_dict[node_uid]['children']:
        if not visited_flags[child_node_uid]:
            _recursive_topological_sort(child_node_uid, graph_dict, stack, visited_flags)
        
    stack.append(node_uid)
    
def topological_sort(graph_dict):
    
    stack = []
    visited_flags = {k:False for k in graph_dict.keys()}

    for node_uid in graph_dict.keys():
        if not visited_flags[node_uid]:
            _recursive_topological_sort(node_uid, graph_dict, stack, visited_flags)
            
    sorted_graph_dict = {}
    while len(stack)>0:
        node_uid = stack.pop()
        sorted_graph_dict[node_uid] = graph_dict[node_uid]
    
    return sorted_graph_dict
            
# def add_to_module_global_namespace(method, global_dict):

#     method.__globals__.update(global_dict)



from .node import DataNode
from .node import OperationNode
from .utils import ExtendedRef
from .utils import view_full
from .utils import view_summary
from .utils import save_graph_image
from .utils import contains_return_statement
from .utils import topological_sort
# from .utils import add_to_module_global_namespace

from collections import defaultdict
from collections import Iterable
import sys
import copy

MAX_INTEGER = sys.maxsize  


class GraphBuilder():
    
    def __init__(self, persist=False, verbose=False, alias=None, inside_pandasUDF=None):#, shared_args=dict()):

        self.graph_alias = alias or "graph"
        self.graph_uid = "{}_{}".format(self.graph_alias, id(self))

        self.persist = persist
        self.verbose = verbose
        
        self.node_count = 0
        self.strong_ref_dict = {}
        self.graph_dict = defaultdict(dict)

        self.default_graph_attributes = {
            'data_node_fontsize': 10, 
            'data_node_shape': 'box',
            'data_node_color': None,
            'op_node_fontsize': 12,
            'op_node_shape': 'box',
            'op_node_color': 'white',
            'graph_ranksep': 0.475,
            'graph_node_fontsize': 12.85,
            'graph_node_shape': 'box3d',
            'graph_node_color': 'white',
            'graph_node_shapesize': 0.574,
            'persist_record_shape': True
        }
        self.user_defined_graph_attributes = None

        self.inside_pandasUDF = inside_pandasUDF

        # if not isinstance(shared_args, dict):
        #     raise TypeError("shared_args must be a dictionary. Instead received {}".format(shared_args))
        # self.shared_args = shared_args
    
    def add(self, func, method_alias=None, output_alias=None, n_out=1, persist=False, rank=None, color=None, shape=None, fontsize=None):

        # add_to_module_global_namespace(func, self.shared_args)
        
        self.func = func
        self.method_alias = method_alias
        self.output_alias = output_alias

        self.n_out = n_out
        self.func_persist = persist

        if ( not isinstance(self.output_alias, list) and not isinstance(self.output_alias, tuple) ):
            self.output_alias = [self.output_alias]
        if not output_alias:
            self.output_alias *= n_out

        if rank is None:
            self.rank = MAX_INTEGER
        else:
            self.rank = rank

        self.color = color
        self.shape = shape
        self.fontsize = fontsize
        
        return self
    
    def __call__(self, *args):

        # Create/update the graph using doubly linked list data structure

        op_node_uid = '{}_{}'.format(self.method_alias or self.func.__name__, self.node_count)

        # create op node to hold the input function method
        self.strong_ref_dict[op_node_uid] = OperationNode(
            graph_uid=self.graph_uid, 
            graph_alias=self.graph_alias, 
            node_uid=op_node_uid, 
            function=self.func, 
            n_out=self.n_out, 
            verbose=self.verbose, 
            alias=self.method_alias,
            graph_dict=self.graph_dict)
        op_node_weak_ref = ExtendedRef(self.strong_ref_dict[op_node_uid])
        self.node_count += 1  

        # create edge: parent data_nodes <--> op_node
        for i, inp in enumerate(args):
            
            # if the inp is actually another node:
            if isinstance(inp, ExtendedRef) and isinstance(inp(), DataNode):

                # inp is already (a weak reference to) a data node

                # add the weak ref to the parent data node to the op node's references
                op_node_weak_ref().parent_node_weak_refs.append(inp)  

                # add the weak reference of the current op node to the parent data node's references 
                inp().child_node_weak_refs.append(op_node_weak_ref)

            # if the inp is raw data, and not another data node
            else:

                parent_data_node_uid = 'data_{}'.format(self.node_count)  # raw data has no alias

                # if inp is raw value, we need to persist since we can't re-compute it from graph
                # create a new data node to hold this raw data
                persist_this_node = True

                # define a new data node that will hold the raw input data
                self.strong_ref_dict[parent_data_node_uid] = DataNode(
                    graph_uid=self.graph_uid, 
                    graph_alias=self.graph_alias, 
                    node_uid=parent_data_node_uid, 
                    persist=persist_this_node, 
                    verbose=self.verbose,
                    graph_dict=self.graph_dict)
                data_node_weak_ref = ExtendedRef(self.strong_ref_dict[parent_data_node_uid])
                self.node_count += 1

                # set the value of the new data node with the input inp value
                data_node_weak_ref().set_value(inp)
                
                # add the weak ref to the parent data node (newly created) to the op node's references
                op_node_weak_ref().parent_node_weak_refs.append(data_node_weak_ref)

                # add the weak reference of the current op node to the parent data node's references 
                data_node_weak_ref().child_node_weak_refs.append(op_node_weak_ref)

        # create edge: op_node <--> n_out children data_node(s) 
        for i in range(self.n_out):

            # if the current method has no return statement, we do not want to create a child data node
            if not (self.inside_pandasUDF or contains_return_statement(self.func)):
                continue

            child_data_node_uid = '{}_{}'.format(self.output_alias[i] or 'data', self.node_count)

            persist_this_node = self.persist or self.func_persist

            # define a new data node that will hold the result of the current op_node's computations
            self.strong_ref_dict[child_data_node_uid] = DataNode(
                graph_uid=self.graph_uid, 
                graph_alias=self.graph_alias,
                node_uid=child_data_node_uid, 
                verbose=self.verbose, 
                persist=persist_this_node, 
                alias=self.output_alias[i],
                graph_dict=self.graph_dict)
            data_node_weak_ref = ExtendedRef(self.strong_ref_dict[child_data_node_uid])
            self.node_count += 1  

            # give the (weak) reference of the above children data_node to the op_node so that
            # the current op_node points to the data_node
            op_node_weak_ref().child_node_weak_refs.append(data_node_weak_ref)

            # add the weak reference of the current op node to the child data node's references
            data_node_weak_ref().parent_node_weak_refs.append(op_node_weak_ref)

        # Create/update the graph_dict for visualization

        # define the graph visualization attributes of the current op node
        node_graph_attributes_dict = {'rank': self.rank, 
                                      'color': self.color, 
                                      'shape': self.shape,
                                      'fontsize': self.fontsize,
                                      'shapesize': None}

        op_node_properties_dict = {'children': [], 
                                   'parents': [], 
                                   'type': 'operation', 
                                   'is_persisted': False,
                                   'data_dim': '',
                                   'alias': op_node_weak_ref().alias,
                                   'node_uid': op_node_weak_ref().node_uid,
                                   'graph_alias': self.graph_alias,
                                   'graph_uid': op_node_weak_ref().graph_uid,
                                   'attributes': node_graph_attributes_dict}

        # add the op node visualization attributes to the graph_dict
        self.graph_dict[op_node_uid] = op_node_properties_dict

        # create edge: op_node <--> parent data_nodes
        for parent_data_node_weak_ref in op_node_weak_ref().parent_node_weak_refs:

            # if the parent data node comes from a different graph
            # the second condition is there to prevent detached op node when the previous
            # graph result is used in more than 1 place.
            if parent_data_node_weak_ref().graph_uid != self.graph_uid:

                _ext_node_uid = "{} from {}".format(
                    parent_data_node_weak_ref().node_uid, 
                    parent_data_node_weak_ref().graph_uid)

                if _ext_node_uid not in self.graph_dict:

                    node_graph_attributes_dict = {'rank': MAX_INTEGER, 
                                                  'color': "red", 
                                                  'shape': None,
                                                  'fontsize': self.fontsize,
                                                  'shapesize': None}

                    node_properties_dict = {'children': [], 
                                            'parents': [],  
                                            'type': 'data',
                                            'is_persisted': parent_data_node_weak_ref().is_persisted(),
                                            'data_dim': parent_data_node_weak_ref().get_persisted_data_dim_as_str(),
                                            'alias': parent_data_node_weak_ref().alias,
                                            'node_uid': _ext_node_uid,
                                            'graph_alias': parent_data_node_weak_ref().graph_alias,
                                            'graph_uid': parent_data_node_weak_ref().graph_uid,
                                            'attributes': node_graph_attributes_dict}

                    # add the parent data node visualization attributes to the graph_dict
                    self.graph_dict[_ext_node_uid] = node_properties_dict

                # make the parent data node point to the op node
                self.graph_dict[_ext_node_uid]['children'].append(op_node_weak_ref().node_uid)

                # make the op node point to the parent data node
                self.graph_dict[op_node_weak_ref().node_uid]['parents'].append(_ext_node_uid)

                continue

            # if the parent data node is not already part of the graph_dict
            elif parent_data_node_weak_ref().node_uid not in self.graph_dict:

                node_graph_attributes_dict = {'rank': MAX_INTEGER, 
                                              'color': None, 
                                              'shape': None,
                                              'fontsize': self.fontsize,
                                              'shapesize': None}

                if parent_data_node_weak_ref().has_value():
                    data_dim = parent_data_node_weak_ref().get_persisted_data_dim_as_str()
                else:
                    data_dim = ''

                node_properties_dict = {'children': [], 
                                        'parents': [], 
                                        'type': 'data',
                                        'is_persisted': parent_data_node_weak_ref().is_persisted(),
                                        'data_dim': data_dim,
                                        'alias': parent_data_node_weak_ref().alias,
                                        'node_uid': parent_data_node_weak_ref().node_uid,
                                        'graph_alias': self.graph_alias,
                                        'graph_uid': self.graph_uid,
                                        'attributes': node_graph_attributes_dict}

                # add the parent data node visualization attributes to the graph_dict
                self.graph_dict[parent_data_node_weak_ref().node_uid] = node_properties_dict

            # make the parent data node point to the op node
            self.graph_dict[parent_data_node_weak_ref().node_uid]['children'].append(op_node_weak_ref().node_uid)

            # make the op node point to the parent data node
            self.graph_dict[op_node_weak_ref().node_uid]['parents'].append(parent_data_node_weak_ref().node_uid)

        # create edge: op_node <--> n_out children data_node(s)
        for child_data_node_weak_ref in op_node_weak_ref().child_node_weak_refs:

            # if the child data node is not already part of the graph_dict
            if child_data_node_weak_ref().node_uid not in self.graph_dict:

                node_graph_attributes_dict = {'rank': MAX_INTEGER, 
                                              'color': None, 
                                              'shape': None,
                                              'fontsize': self.fontsize,
                                              'shapesize': None}

                node_properties_dict = {'children': [], 
                                        'parents': [], 
                                        'type': 'data', 
                                        'is_persisted': child_data_node_weak_ref().is_persisted(),
                                        'data_dim': '',
                                        'alias': child_data_node_weak_ref().alias,
                                        'node_uid': child_data_node_weak_ref().node_uid,
                                        'graph_alias': self.graph_alias,
                                        'graph_uid': self.graph_uid,
                                        'attributes': node_graph_attributes_dict}

                # add the child data node visualization attributes to the graph_dict
                self.graph_dict[child_data_node_weak_ref().node_uid] = node_properties_dict

            # make the child data node point to the op node
            self.graph_dict[child_data_node_weak_ref().node_uid]['parents'].append(op_node_weak_ref().node_uid)

            # make the op node point to the child data node
            self.graph_dict[op_node_weak_ref().node_uid]['children'].append(child_data_node_weak_ref().node_uid)
                
        if self.n_out > 1:
            return op_node_weak_ref().child_node_weak_refs
        else:
            if len(op_node_weak_ref().child_node_weak_refs) == 0:
                return None
            else:
                return op_node_weak_ref().child_node_weak_refs[0]

    def run(self, *args):    

        requested_data_node_uids = [elem().node_uid for elem in args]
        
        requested_data_nodes = [(k, v) for k, v in self.strong_ref_dict.items() 
                                     if v.node_uid in requested_data_node_uids]
        
        for k, v in requested_data_nodes:
            v.persist()
        
        op_nodes = [(k, v) for k, v in self.strong_ref_dict.items() if v.node_type == 'operation']

        for k, v in op_nodes:
            v.activate()
        
        for k, v in op_nodes:
            v.run()

        # because op node calls get only on the parent data node, 
        # and directly assigns the current function result to child node
        # invoke get() on the last op node's child data node to make sure the 
        # necessary persistance occurs.
        for child_data_node_weak_ref in v.child_node_weak_refs:
            child_data_node_weak_ref().get()
        
        if len(args) == 1:
            return args[0].get()
        else:
            return [arg.get() for arg in args]

    def remove(self, n=1):

        op_nodes = [(k, v) for k, v in self.strong_ref_dict.items() if v.node_type == 'operation']
        rev_op_nodes = op_nodes[::-1]
        rev_op_nodes = rev_op_nodes[:n]

        for k, v in rev_op_nodes:

            # first get all the children nodes of the op node
            child_data_node_uids = self.graph_dict[k]['children']
            
            # remove the child data nodes 
            # no need to unlink them from its parent op node since
            # all data node has only one parent op node (i.e. k)
            for child_data_node_uid in child_data_node_uids:
                
                # remove from graph dict
                self.graph_dict.pop(child_data_node_uid)
                
                # remove the strong reference from memory
                del self.strong_ref_dict[child_data_node_uid]

                self.node_count -= 1
            
            # get all the parent data nodes of the op node
            parent_data_node_uids = self.graph_dict[k]['parents']
            
            # unlink the current op node (i.e. k) from its parent 
            # data nodes
            for parent_data_node_uid in parent_data_node_uids:
                
                # if the parent data node is raw input, remove it
                if len(self.graph_dict[parent_data_node_uid]['parents']) == 0:
                    
                    # remove from graph dict
                    self.graph_dict.pop(parent_data_node_uid)
                    
                    # remove the strong reference from memory
                    # if the parent data node was from a different graph, 
                    # we don't want to release its memory
                    if parent_data_node_uid in self.strong_ref_dict:
                        del self.strong_ref_dict[parent_data_node_uid]

                    self.node_count -= 1
                    
                    continue
                    
                # unlink from parent data node in graph dict
                old_child_node_uids = self.graph_dict[parent_data_node_uid]['children']
                new_child_node_uids = [elem for elem in old_child_node_uids if elem != k]
                self.graph_dict[parent_data_node_uid]['children'] = new_child_node_uids
                
                # unlink from parent data node in memory
                self.strong_ref_dict[parent_data_node_uid].remove_child_node(self.strong_ref_dict[parent_data_node_uid])
                self.strong_ref_dict[parent_data_node_uid].remove_child_node(v)
                
            # remove the op node from graph dict
            self.graph_dict.pop(k)
            
            # remove the strong reference to it
            del self.strong_ref_dict[k]

            self.node_count -= 1
            
    @property
    def graph_attributes(self):

        if self.user_defined_graph_attributes:
            graph_attributes = copy.copy(self.default_graph_attributes)
            graph_attributes.update(self.user_defined_graph_attributes)
        else:
            graph_attributes = copy.copy(self.default_graph_attributes)

        return graph_attributes

    def _graph_attributes(self):

        return {k: str(v) for k, v in self.graph_attributes.items()}

    def preprocess_graph_dict(self):
        """
        1. support for multi graph
        - need to create a copy of graph dict so that we can give it an appendage of graph node
        """
        graph_dict = copy.deepcopy(self.graph_dict)
        external_graphs_dict = {k: v for k, v in graph_dict.items() if v['graph_uid']!=self.graph_uid}

        for k, v in external_graphs_dict.items():

            node_graph_attributes_dict = {'rank': MAX_INTEGER, 
                                          'color': self.graph_attributes['graph_node_color'], 
                                          'shape': self.graph_attributes['graph_node_shape'],
                                          'fontsize': self.graph_attributes['graph_node_fontsize'],
                                          'shapesize': self.graph_attributes['graph_node_shapesize']}

            op_node_properties_dict = {'children': [k], 
                                       'parents': [], 
                                       'type': 'operation', 
                                       'alias': v['graph_alias'],
                                       'node_uid': k + "_ghost",
                                       'graph_alias': v['graph_alias'],
                                       'graph_uid': v['graph_uid'],
                                       'attributes': node_graph_attributes_dict}

            v['parents'].append(k+"_ghost")
            graph_dict[k+"_ghost"] = op_node_properties_dict

        sorted_graph_dict = topological_sort(graph_dict)
        return sorted_graph_dict

    def view(self, summary=True, graph_attributes=None):

        if graph_attributes:  # need validity check here
            self.user_defined_graph_attributes = graph_attributes
        else:
            self.user_defined_graph_attributes = None  # I don't like this

        preprocessed_graph_dict = self.preprocess_graph_dict()
        
        if summary:
            return view_summary(preprocessed_graph_dict, self._graph_attributes(), verbose=self.verbose, current_graph_uid=self.graph_uid)
        else:
            return view_full(preprocessed_graph_dict, self._graph_attributes(), verbose=self.verbose, current_graph_uid=self.graph_uid)

    def save_view(self, summary=True, graph_attributes=None, dirpath=None, filename='digraph', fileformat='png'):

        if fileformat not in ['pdf', 'png']:
            raise TypeError("Expected fileformat to be 'pdf' or 'png', but instead "
                            "got {}".format(fileformat))

        graph = self.view(summary, graph_attributes)
        img_filepath = save_graph_image(graph, dirpath, filename, fileformat)
        return img_filepath



from setuptools import setup, find_packages


import pyflow

VERSION = pyflow.__version__

with open("README.rst", "r") as fh:
    long_description = fh.read()

setup(
	name="pyflow-viz", 
	version=VERSION,
	author="Jay Kim",
	description="Lazy computation directed acyclic graph builder",
	long_description=long_description,
	long_description_content_type="text/x-rst",
	url=None,
	license="DSB 3-clause",
	packages=find_packages(),
	install_requires=["graphviz>=0.13.2"]
	)
